/* Copyright 2011, Unpublished Work of Technologic Systems
 * All Rights Reserved.
 *
 * THIS WORK IS AN UNPUBLISHED WORK AND CONTAINS CONFIDENTIAL,
 * PROPRIETARY AND TRADE SECRET INFORMATION OF TECHNOLOGIC SYSTEMS.
 * ACCESS TO THIS WORK IS RESTRICTED TO (I) TECHNOLOGIC SYSTEMS 
 * EMPLOYEES WHO HAVE A NEED TO KNOW TO PERFORM TASKS WITHIN THE SCOPE
 * OF THEIR ASSIGNMENTS AND (II) ENTITIES OTHER THAN TECHNOLOGIC
 * SYSTEMS WHO HAVE ENTERED INTO APPROPRIATE LICENSE AGREEMENTS.  NO
 * PART OF THIS WORK MAY BE USED, PRACTICED, PERFORMED, COPIED, 
 * DISTRIBUTED, REVISED, MODIFIED, TRANSLATED, ABRIDGED, CONDENSED, 
 * EXPANDED, COLLECTED, COMPILED, LINKED, RECAST, TRANSFORMED, ADAPTED
 * IN ANY FORM OR BY ANY MEANS, MANUAL, MECHANICAL, CHEMICAL, 
 * ELECTRICAL, ELECTRONIC, OPTICAL, BIOLOGICAL, OR OTHERWISE WITHOUT
 * THE PRIOR WRITTEN PERMISSION AND CONSENT OF TECHNOLOGIC SYSTEMS.
 * ANY USE OR EXPLOITATION OF THIS WORK WITHOUT THE PRIOR WRITTEN
 * CONSENT OF TECHNOLOGIC SYSTEMS COULD SUBJECT THE PERPETRATOR TO
 * CRIMINAL AND CIVIL LIABILITY.
 */

/* This module implements a DIO expanding scan cycle that is specific to the
 * TS-7520 and similar boards.  It presents up to 128 outputs and 64 inputs to
 * internal logic as if they were direct DIOs.
 *
 * The outputs are all via the CPLD.  The CPLD latches output values on each
 * rising edge of scan_clk.  This module provides output data setup and hold
 * times of 2*clk_i period.
 *
 * Inputs come from 7 mux chips, and 1 from the CPLD which acts as an eighth
 * mux chip for this purpose.  The 3 bits of scan address are presented to the
 * mux chips, giving them 8 clk_i cycles to set up their outputs.
 *
 * All inputs are slow, but this module may need to be concerned with 
 * performance on outputs.  For that reason there is no input sampling during
 * higher address output cycles.  There are a maximum of 76 output bits, so
 * we have 19 output cycles.  Inputs are sampled on every other output cycle
 * for the first 16 output cycles.
 *
 * It may be necessary to drive an output with a PWM.  In that case the base
 * frequency of the PWM should be informed by the output frequency of this
 * module.  To that end, this module outputs a counter which is synchronized
 * to the scan cycle and is intended to be used in a PWM comparator.  Any tap
 * from this counter is valid for this purpose.  The nominal frequency will be
 * (75/76)MHz.  This allows for a 10 bit 1kHz PWM, an 8 bit 4kHz PWM, etc.
 * If better PWM resolution/speed is required, we have to redesign the scan
 * cycle with that in mind.
 * 
 * MUX's have a max delay of 62ns
 */

module scan_cycle (
  input clk_i,
  input rst_i,

  input [127:0] outputs_i,
  output [71:0] inputs_o,

  output [4:0] scan_add_o,
  output scan_clk_o,
  output [3:0] scan_dout_o,
  input [8:0] scan_din_i
);

parameter add_max = 5'd31; // 31 addresses = 124 outputs possible with this scheme

reg [71:0] dio;
assign inputs_o = dio;

reg [4:0] add;
reg [1:0] scan_clk;

assign scan_dout_o = outputs_i[4*add+:4];
assign scan_add_o = add;
assign scan_clk_o = scan_clk[1];
  

always @(posedge clk_i or posedge rst_i) begin
  if (rst_i) begin
    dio <= 72'd0;
    add <= 5'd0;
    scan_clk <= 2'b00;
  end else begin
    scan_clk <= scan_clk + 1'b1;
    if (scan_clk == 2'b11) begin
      // update outputs with scan_clk falling edge:
      add <= add + 1'b1;
      if (add >= add_max) add <= 5'd0;

      // read inputs once per 8 clocks:
      if (add[0]) dio[add[3:1]*9+:9] <= scan_din_i;
    end
  end
end
endmodule

