/* Copyright 2011, Unpublished Work of Technologic Systems
 * All Rights Reserved.
 *
 * THIS WORK IS AN UNPUBLISHED WORK AND CONTAINS CONFIDENTIAL,
 * PROPRIETARY AND TRADE SECRET INFORMATION OF TECHNOLOGIC SYSTEMS.
 * ACCESS TO THIS WORK IS RESTRICTED TO (I) TECHNOLOGIC SYSTEMS 
 * EMPLOYEES WHO HAVE A NEED TO KNOW TO PERFORM TASKS WITHIN THE SCOPE
 * OF THEIR ASSIGNMENTS AND (II) ENTITIES OTHER THAN TECHNOLOGIC
 * SYSTEMS WHO HAVE ENTERED INTO APPROPRIATE LICENSE AGREEMENTS.  NO
 * PART OF THIS WORK MAY BE USED, PRACTICED, PERFORMED, COPIED, 
 * DISTRIBUTED, REVISED, MODIFIED, TRANSLATED, ABRIDGED, CONDENSED, 
 * EXPANDED, COLLECTED, COMPILED, LINKED, RECAST, TRANSFORMED, ADAPTED
 * IN ANY FORM OR BY ANY MEANS, MANUAL, MECHANICAL, CHEMICAL, 
 * ELECTRICAL, ELECTRONIC, OPTICAL, BIOLOGICAL, OR OTHERWISE WITHOUT
 * THE PRIOR WRITTEN PERMISSION AND CONSENT OF TECHNOLOGIC SYSTEMS.
 * ANY USE OR EXPLOITATION OF THIS WORK WITHOUT THE PRIOR WRITTEN
 * CONSENT OF TECHNOLOGIC SYSTEMS COULD SUBJECT THE PERPETRATOR TO
 * CRIMINAL AND CIVIL LIABILITY.
 */

/* This core is for generating up to 8 PWM outputs.  Its features:
 *  - 12-bits of duty cycle (0 - 0%,  0xfff - 100%)
 *  - 2 external frequency trigger inputs
 *  - 2 internal frequency generators (each PWM chan picks one or the other)
 *  - Only one 16-bit write-only register conserves address decode
 *  - Various parameter "knobs" can save LUTs when features unneeded
 *  - PWM frequency can go as low as 10Hz
 *  - Small area (6 chans + 1 ext trig + 1 freq generator = 250LUTs)
 *  - Optional 3-phase VFD power inverter at 12Khz PWM and 0-127Hz sinusoids
 *
 * The core is controlled via opcodes written to its one write-only WISHBONE
 * register. There are actually 8 internal 13 bit registers that can be 
 * written but the basic format of the WISHBONE write is as follows:
 *
 *   bits 15-13: internal write-only register number
 *   bits 12-0: internal write-only register data
 * 
 * Internal write-only registers:
 *
 *   0: PWM control channel #0
 *     bit 15-13: must be 0
 *     bit 12: frequency select
 *     bit 11-0: duty cycle percent (0 - 0%, 0xfff - 100%)
 *   1: PWM control channel #1
 *     bit 15-13: must be 1
 *     bit 12: frequency select
 *     bit 11-0: duty cycle percent (0 - 0%, 0xfff - 100%)
 *   2: PWM control channel #2 ...
 *   3: PWM control channel #3 ...
 *   4: PWM control channel #4 ...
 *   5: PWM control channel #5 ...
 *   6: PWM frequency generator #0 (or PWM channel #6)
 *     bit 15-13: must be 6
 *     bit 12: period units (0 - 1 us, 1 - 100 us)
 *     bit 11-0: frequency period (multiplied by units)
 *   7: PWM frequency generator #1 (or PWM channel #7)
 *     bit 15-13: must be 7
 *     bit 12: period units (0 - 1 us, 1 - 100 us)
 *     bit 11-0: frequency period (multiplied by units)
 *
 * The frequency select (bit 12) of the PWM control channel #0-7 regs
 * selects which frequency to use and it depends on the parameters used 
 * during synthesis.  For instance, for bit == 0, precedence is interpreted
 * as follows:
 *   ... if hardcoded_freq0 is non-zero, then use that frequency, else ... 
 *   ... if use_exttrig0 is set, then use module port exttrig0_i, else ...
 *   ... use internal frequency generator #0
 *
 * Note that it is not possible to use all 8 channels and use any 
 * of the internal frequency generators.  The design will synthesize, but not
 * make much sense.
 *
 *
 * 3-phase VFD power inverter option:
 * ==================================
 *
 * This core can also be used to generate VFD (variable frequency drive)
 * 3-phase sinusoid PWM waveforms.  This uses a hardcoded 12Khz PWM frequency
 * where duty-cycle is continuously varied to output 3 sinusoid waveforms
 * 120 degress out of phase with each other.  The sinusoid frequency can be
 * from 0-127Hz with 13-bits of precision.  The amplitude can also be varied
 * from 0-100% with 12-bits of precision.  It is possible to reverse the motor
 * which is accomplished by swapping phases B and C.
 * 
 * When enabled, registers #5 and #6 have new meanings:
 *   5: VFD sinusoid amplitude (or PWM channel #5)
 *     bit 15-13: must be 5
 *     bit 12: reverse enable (0 - normal, 1 - reversed)
 *     bit 11-0: amplitude percent (0 - 0%, 0xfff - 100%)
 *   6: VFD sinusoid frequency (or PWM channel #6 / PWM frequency generator #0)
 *     bit 15-13: must be 6
 *     bit 12-0: frequency (0 - 0Hz, 0x1fff - ~127Hz)
 * 
 * Precise frequency in Hz is = (1.0 / ((0x80000000 / freq[12:0]) * 30e-9))
 *
 * Note that because the VFD overrides PWM registers #5 and #6, it will not
 * make much sense to have the core 'nchan' parameter any higher than 5.
 *
 * 
 * Compile-time parameters:
 * ========================
 * 
 * nchan: number of channels (default 6)
 *
 * single_freq: set to 1 if its not neccessary to allow users to select
 * between one of two frequencies for each channel.  This hardcodes to use
 * only freq #0.
 *
 * use_exttrig0, use_exttrig1: set either to one to use the external
 * trigger input. This overrides using the internal frequency generator.
 *
 * hardcoded_freq0, hardcoded_freq1: set either to non-zero to use a
 * hardcoded frequency input.  Expects a frequency in Hz.  This overrides
 * using the internal frequency generator AND using the external trigger.
 *
 * wb_freq: must be frequency in Hz of wb_clk_i input. (default 75Mhz)
 *
 * vfd_opt: set to 1 to enable the 3-phase VFD core.  
 *
 * nonpwm_opt: set to 1 to remove all PWM capabilities of this core and 
 * round pwm_o outputs to either 100% duty or 0% duty (constant 1 or 0). 
 * This can be used as a simple way to reduce logic resources.
 *
 * The compile-time parameters can be read from software by reading the
 * write-only register:  
 *  
 *   bits 15-8: Magic value "0x75"
 *   bit 7: hardcoded frequency is set for freq#1
 *   bit 6: hardcoded frequency is set for freq#0
 *   bit 5: uses external trigger input#1 for freq#1
 *   bit 4: uses external trigger input#0 for freq#0
 *   bit 3: uses only a single frequency input
 *   bits 2-0: number of PWM channels compiled-in
 *
 */

`undef TESTBENCH
`ifdef TESTBENCH
`timescale 1 ns / 1 ps
`endif
module wb_pwm(
  wb_clk_i,
  wb_rst_i,

  wb_cyc_i,
  wb_stb_i,
  wb_adr_i,
  wb_dat_i,
  wb_dat_o,
  wb_we_i,
  wb_ack_o,

  pwm_o,
  nonpwm_o,
  vfd_pwm_a_o,
  vfd_pwm_b_o,
  vfd_pwm_c_o,
  vfd_a_o,
  vfd_b_o,
  vfd_c_o,
  exttrig0_i,
  exttrig1_i,
  clk_100mhz_i /* For internal frequency generator */
);
input wb_clk_i;
input wb_rst_i;
input wb_cyc_i;
input wb_stb_i;
input wb_we_i;
input [31:0] wb_adr_i;
input [15:0] wb_dat_i;
output [15:0] wb_dat_o;
output wb_ack_o;

output [7:0] pwm_o, nonpwm_o;
output [1:0] vfd_pwm_a_o, vfd_pwm_b_o, vfd_pwm_c_o;
output [11:0] vfd_a_o, vfd_b_o, vfd_c_o;
input exttrig0_i;
input exttrig1_i;
input clk_100mhz_i;

parameter nchan = 6;
parameter single_freq = 0;
parameter use_exttrig0 = 1; 
parameter use_exttrig1 = 0; 
parameter hardcoded_freq0 = 0;
parameter hardcoded_freq1 = 0;
parameter wb_freq = 75000000; /* Used if a hardcoded_freq is set */
parameter vfd_opt = 1;
parameter nonpwm_opt = 0;

integer hardcoded_period0 = (1.0 / hardcoded_freq0) / (1.0 / wb_freq) - 1.5;
integer hardcoded_period1 = (1.0 / hardcoded_freq1) / (1.0 / wb_freq) - 1.5;

assign wb_dat_o[5:0] = 
  {use_exttrig1[0], use_exttrig0[0], single_freq[0], nchan[2:0]};
assign wb_dat_o[6] = (hardcoded_freq0 > 0);
assign wb_dat_o[7] = (hardcoded_freq1 > 0);
assign wb_dat_o[15:8] = 8'h75;

/* in units of 10 ns, max representable period = 168ms = 5.96hz */
reg [23:0] period0, counter0, period1, counter1;
reg [12:0] regmem[0:7]; /* 0 - 0%, 12'hfff - 100% */
reg [35:0] multiplied;
reg wb_ack, trigsel2, trigsel3, w;
reg [2:0] seq, seq2, seq3;
reg [7:0] pwm, pwm_0pct, pwm_100pct, pwm_rst, counter0_pwms, counter1_pwms;
reg [7:0] nonpwm;
assign pwm_o = nonpwm_opt ? nonpwm : (pwm & ~pwm_0pct | pwm_100pct);
assign nonpwm_o = nonpwm;
assign wb_ack_o = wb_ack;
reg [12:0] trigctl0, trigctl1;
wire inttrig0, inttrig1;

wire trig0 = use_exttrig0[0] ? exttrig0_i : inttrig0;
wire trig1 = use_exttrig1[0] ? exttrig1_i : inttrig1;

integer i;
wire trigsel = single_freq[0] ? 1'b0 : regmem[seq][12];
reg [23:0] multa, multb;
always @(posedge wb_clk_i or posedge wb_rst_i) begin
  if (wb_rst_i) begin
    multa <= 24'd0;
    multb <= 24'd0;
    wb_ack <= 1'b0;
    multiplied <= 36'd0;
    period0 <= hardcoded_freq0 > 0 ? hardcoded_period0 : 24'hffffff;
    counter0 <= 24'd0;
    period1 <= hardcoded_freq1 > 0 ? hardcoded_period1 : 24'hffffff;
    counter1 <= 24'd0;
    pwm <= 8'd0;
    pwm_0pct <= 8'hff;
    pwm_100pct <= 8'd0;
    trigsel2 <= 1'b0;
    trigsel3 <= 1'b0;
    seq <= 3'd0;
    seq2 <= 3'd0;
    seq3 <= 3'd0;
    trigctl0 <= {1'b1, 12'd99}; /* default 100Hz */
    trigctl1 <= {1'b1, 12'd99}; 
    for (i = 0; i < 8; i = i + 1) regmem[i] <= 13'd0;
    counter0_pwms <= 8'd0;
    counter1_pwms <= 8'd0;
    pwm_rst <= 8'd0;
    w <= 1'b0;
    nonpwm <= 8'd0;
  end else begin
    counter0 <= counter0 + 1'b1;
    counter1 <= counter1 + 1'b1;
    wb_ack <= 1'b0;
    if (wb_cyc_i && wb_stb_i && !wb_ack) wb_ack <= 1'b1;
    if (wb_cyc_i && wb_stb_i && wb_we_i && !wb_ack) begin
      w <= !w;
      if (wb_dat_i[11]) nonpwm[wb_dat_i[15:13]] <= 1'b1;
      else nonpwm[wb_dat_i[15:13]] <= 1'b0;
      if (wb_dat_i[15:13] == 3'd7) trigctl1 <= wb_dat_i[12:0];
      if (wb_dat_i[15:13] == 3'd6) trigctl0 <= wb_dat_i[12:0];
      if (wb_dat_i[15:13] < nchan || ((wb_dat_i[15:13] == 3'd5 || 
        wb_dat_i[15:13] == 3'd6) && vfd_opt)) begin
	regmem[wb_dat_i[15:13]] <= wb_dat_i[12:0];
        if (wb_dat_i[11:0] != 12'd0) pwm_0pct[wb_dat_i[15:13]] <= 1'b0;
        else pwm_0pct[wb_dat_i[15:13]] <= 1'b1;
        if (wb_dat_i[11:0] != 12'hfff) pwm_100pct[wb_dat_i[15:13]] <= 1'b0;
        else pwm_100pct[wb_dat_i[15:13]] <= 1'b1;
      end
    end 

    if (seq >= (nchan - 1'b1)) seq <= 3'd0;
    else seq <= seq + 1'b1;
  
    if (hardcoded_freq1 > 0) begin
      if (counter1 >= hardcoded_period1) begin
        counter1 <= 24'd0;
        pwm_rst <= pwm_rst | counter1_pwms;
      end
    end else if (trig1) begin
      counter1 <= 24'd0;
      period1 <= counter1 + 1'b1;
      pwm_rst <= pwm_rst | counter1_pwms;
    end

    if (hardcoded_freq0 > 0) begin
      if (counter0 >= hardcoded_period0) begin
        counter0 <= 24'd0;
        pwm_rst <= pwm_rst | counter0_pwms;
      end
    end else if (trig0) begin
      counter0 <= 24'd0;
      period0 <= counter0 + 1'b1;
      pwm_rst <= pwm_rst | counter0_pwms;
    end

    multa <= {12'd0, regmem[seq][11:0]};
    multb <= (trigsel ? period1 : period0);
    trigsel2 <= trigsel;
    seq2 <= seq;
  
    multiplied <= multa * multb;
    trigsel3 <= trigsel2;
    seq3 <= seq2;

    if (pwm_rst[seq3]) begin
      pwm_rst[seq3] <= 1'b0;
      pwm[seq3] <= 1'b1;
    end
  
    if (multiplied[35:12] <= (trigsel3 ? counter1 : counter0)) begin
      pwm[seq3] <= 1'b0;
      counter1_pwms[seq3] <= trigsel3;
      counter0_pwms[seq3] <= !trigsel3;
    end
  end
end

periodic_trig freqgen0(
  .rst_i(wb_rst_i),
  .clk_i(wb_clk_i),
  .clk_100mhz_i(clk_100mhz_i),
  .timescale_i(trigctl0[12]),
  .count_i({20'd0, trigctl0[11:0]}),
  .trig_o(inttrig0)
);

periodic_trig freqgen1(
  .rst_i(wb_rst_i),
  .clk_i(wb_clk_i),
  .clk_100mhz_i(clk_100mhz_i),
  .timescale_i(trigctl1[12]),
  .count_i({20'd0, trigctl1[11:0]}),
  .trig_o(inttrig1)
);

/* For 3-phase variable frequency drive (VFD) power inverter */
reg [1:0] w_q;
reg [11:0] amplitude;
reg [12:0] freq;
reg reverse;
always @(posedge clk_100mhz_i or posedge wb_rst_i) begin
  if (wb_rst_i) begin
    w_q <= 2'b00;
    amplitude <= 12'd0;
    freq <= 13'd0;
    reverse <= 1'b0;
  end else begin
    w_q <= {w_q[0], w};
    if (w_q[0] != w_q[1]) begin
      amplitude <= regmem[5][11:0];
      freq <= regmem[6][12:0];
      reverse <= regmem[5][12];
    end
  end
end
vfd vfdcore(
  .clk_100mhz_i(clk_100mhz_i && vfd_opt),
  .rst_i(wb_rst_i || !vfd_opt),
  .a_pwm_o(vfd_pwm_a_o),
  .b_pwm_o(vfd_pwm_b_o),
  .c_pwm_o(vfd_pwm_c_o),
  .a_o(vfd_a_o),
  .b_o(vfd_b_o),
  .c_o(vfd_c_o),
  .amplitude_i(amplitude),
  .freq_i(freq),
  .reverse_i(reverse)
);

endmodule

module periodic_trig(
  rst_i,
  clk_100mhz_i,
  timescale_i, /* 0 - 1 microseconds, 1 - 100 microseconds */
  count_i,

  clk_i,
  trig_o
);

input rst_i;
input clk_100mhz_i;
input timescale_i;
input [31:0] count_i;
input clk_i;
output trig_o;

parameter nbits = 12;

localparam N = nbits - 1;

reg [13:0] ts;
reg scale;
reg q;
always @(posedge clk_100mhz_i or posedge rst_i) begin
  if (rst_i) begin
    ts <= 14'd0;
    scale <= 1'b0;
    q <= 1'b0;
  end else begin
    scale <= timescale_i; /* ...so it can be from other clock domain */

    ts <= ts + 1'b1;
    if (ts == (scale ? 14'd9999 : 14'd99)) begin
      ts <= 14'd0;
      q <= !q;
    end
  end
end

reg [1:0] sync;
reg [31:0] timer;
reg trig;
assign trig_o = trig;
always @(posedge clk_i or posedge rst_i) begin
  if (rst_i) begin
    sync <= 2'd0;
    timer <= 32'd0;
    trig <= 1'b0;
  end else begin
    trig <= 1'b0;
    sync <= {sync[0], q};
    if (sync[0] != sync[1]) begin
      if (timer[N:0] >= count_i[N:0]) begin
        trig <= 1'b1;
        timer <= 32'd0;
      end else timer <= timer + 1'b1;
    end
  end
end

endmodule


module vfd (
  clk_100mhz_i,
  rst_i,
  a_o,
  b_o,
  c_o,
  a_pwm_o,
  b_pwm_o,
  c_pwm_o,
  amplitude_i,
  freq_i,
  reverse_i
);

input clk_100mhz_i, rst_i;
output [11:0] a_o, b_o, c_o;
output [1:0] a_pwm_o, b_pwm_o, c_pwm_o;
input [11:0] amplitude_i; /* 0(0%) - 4095(100%) */
input [12:0] freq_i; /* 0(0Hz) - 8191(~127Hz) */
input reverse_i;

/* Sinusoid frequency (in Hz) = (1.0 / ((0x80000000 / freq_i) * 30e-9)) */
/* PWM frequency = 12.21Khz */

parameter deadtime_ns = 400;

reg [15:0] ddstbl [0:1023] /* synthesis syn_romstyle="block_rom" */;
initial $readmemh("./ddstbl.dat", ddstbl);
reg [16:0] ddsout;
reg [28:0] mult;
reg [30:0] pos, pos_q;
reg [1:0] state;
reg [11:0] a, b, c;
reg [12:0] tricnt;
reg [9:0] ddsadr;
wire [11:0] triangle = tricnt[12] ? ~tricnt[11:0] : tricnt[11:0];
reg a_pwm, b_pwm, c_pwm, delta_a, delta_b, delta_c;
reg a_pwm_dead, b_pwm_dead, c_pwm_dead;
assign a_o = a;
assign b_o = reverse_i ? c : b;
assign c_o = reverse_i ? b : c;
assign a_pwm_o[0] = (a_pwm & !a_pwm_dead);
assign b_pwm_o[0] = reverse_i ? (c_pwm & !c_pwm_dead): (b_pwm & !b_pwm_dead);
assign c_pwm_o[0] = reverse_i ? (b_pwm & !b_pwm_dead): (c_pwm & !c_pwm_dead);
assign a_pwm_o[1] = (!a_pwm & !a_pwm_dead);
assign b_pwm_o[1] = reverse_i ? (!c_pwm & !c_pwm_dead): (!b_pwm & !b_pwm_dead);
assign c_pwm_o[1] = reverse_i ? (!b_pwm & !b_pwm_dead): (!c_pwm & !c_pwm_dead);
reg [7:0] deadtimer_a, deadtimer_b, deadtimer_c;
always @(posedge clk_100mhz_i or posedge rst_i) begin
  if (rst_i) begin
    pos <= 31'd0;
    pos_q <= 31'd0;
    state <= 2'd0;
    ddsout <= 17'd0;
    ddsadr <= 10'd0;
    mult <= 28'd0;
    tricnt <= 13'd0;
    a_pwm <= 1'b0;
    b_pwm <= 1'b0;
    c_pwm <= 1'b0;
    delta_a <= 1'b0;
    delta_b <= 1'b0;
    delta_c <= 1'b0;
    a_pwm_dead <= 1'b0;
    b_pwm_dead <= 1'b0;
    c_pwm_dead <= 1'b0;
    a <= 12'd0;
    b <= 12'd0;
    c <= 12'd0;
    deadtimer_a <= 8'd0;
    deadtimer_b <= 8'd0;
    deadtimer_c <= 8'd0;
  end else begin
    /* variable frequency DDS lookup and amplitude scaling */
    case (state)
    2'd0: begin
      a <= mult[28:17];
      pos <= pos + 31'h2aaaaaab;
      state <= 2'd1;
    end
    2'd1: begin
      b <= mult[28:17];
      pos <= pos + 31'h2aaaaaab;
      state <= 2'd2;
    end
    2'd2: begin
      c <= mult[28:17];
      pos <= pos + 31'h2aaaaaaa + freq_i;
      state <= 2'd0;
    end
    endcase
    ddsadr <= pos[29] ? ~pos[28:19] : pos[28:19];
    pos_q <= pos;
    ddsout <= {1'b1, ddstbl[ddsadr]} ^ (pos_q[30] ? ~0 : 0);
    mult <= ddsout[16:0] * amplitude_i;

    /* Triangle wave comparator for PWM generation */
    tricnt <= tricnt + 1'b1;
    if (triangle < a) a_pwm <= 1'b1; else a_pwm <= 1'b0;
    if (triangle < b) b_pwm <= 1'b1; else b_pwm <= 1'b0;
    if (triangle < c) c_pwm <= 1'b1; else c_pwm <= 1'b0;

    /* Dead time counter logic (basically disabled if deadtime_ns == 0) */
    if (triangle < a) delta_a <= !a_pwm; else delta_a <= a_pwm;
    if (triangle < b) delta_b <= !b_pwm; else delta_b <= b_pwm;
    if (triangle < c) delta_c <= !c_pwm; else delta_c <= c_pwm;

    if (delta_a) begin
      deadtimer_a <= 8'd0;
      if (deadtime_ns / 10 > 0) a_pwm_dead <= 1'b1;
    end
    if (a_pwm_dead) begin
      deadtimer_a <= deadtimer_a + 1'b1; 
      if (deadtimer_a >= deadtime_ns / 10) a_pwm_dead <= 1'b0;
    end

    if (delta_b) begin
      deadtimer_b <= 8'd0;
      if (deadtime_ns / 10 > 0) b_pwm_dead <= 1'b1;
    end
    if (b_pwm_dead) begin
      deadtimer_b <= deadtimer_b + 1'b1; 
      if (deadtimer_b >= deadtime_ns / 10) b_pwm_dead <= 1'b0;
    end

    if (delta_c) begin
      deadtimer_c <= 8'd0;
      if (deadtime_ns / 10 > 0) c_pwm_dead <= 1'b1;
    end
    if (c_pwm_dead) begin
      deadtimer_c <= deadtimer_c + 1'b1; 
      if (deadtimer_c >= deadtime_ns / 10) c_pwm_dead <= 1'b0;
    end

  end
end
endmodule


`ifdef TESTBENCH
module test();

reg clk, rst, cyc, stb, trig0, trig1;
reg clk_100mhz;
reg [15:0] dat;
wire ack;
wire [7:0] pwm;
wb_pwm uut(
  .wb_clk_i(clk),
  .wb_rst_i(rst),
  .wb_cyc_i(cyc),
  .wb_stb_i(stb),
  .wb_dat_i(dat),
  .wb_we_i(1'b1),
  .wb_ack_o(ack),

  .exttrig0_i(trig0),
  .exttrig1_i(trig1),
  .pwm_o(pwm),
  .clk_100mhz_i(clk_100mhz)
);

task wb_write16;
  input [15:0] data_i;
  integer i;
  reg [7:0] clocks;
begin
  clocks = 8'd0;
  cyc = 1'b1;
  stb = 1'b1;
  dat = data_i;
  @(posedge clk);
  while (!ack) begin
    @(posedge clk);
    clocks = clocks + 1'b1;
  end
  cyc = 1'b0; stb = 1'b0;
  $display("clocks %d", clocks);
end
endtask

always #7 clk = !clk;
always #5 clk_100mhz = !clk_100mhz;

reg [7:0] counter = 8'd0;
always @(posedge clk) begin
  trig0 <= 1'b0;
  if (counter == 8'd99) begin
    trig0 <= 1'b1;
    counter <= 8'd0;
  end else counter <= counter + 1'b1;
end

initial begin
  $dumpfile("dump.vcd");
  $dumpvars(3, test);
  clk_100mhz = 0; clk = 0; rst = 1; cyc = 0; stb = 0; dat = 0;
  trig1 = 0;
  #100;
  rst = 0;
  wb_write16(16'h0800); /* Chan 0 = 50% */
  wb_write16(16'h2fff); /* Chan 1 = 100% */
  wb_write16(16'h4400); /* Chan 2 = 25% */
  #8800;
  wb_write16(16'h4c00); /* Chan 2 = 75% */
  #10000;
  $finish;
end
  
endmodule
`endif
